import typing

from keras import backend
from keras import layers

from kimm.utils import make_divisible


def apply_activation(
    inputs, activation: typing.Optional[str] = None, name: str = "activation"
):
    x = inputs
    if activation is not None:
        if isinstance(activation, str):
            x = layers.Activation(activation, name=name)(x)
        elif isinstance(activation, layers.Layer):
            x = activation(x)
        else:
            NotImplementedError(
                f"Unsupported activation type: {type(activation)}"
            )
    return x


def apply_conv2d_block(
    inputs,
    filters: typing.Optional[int] = None,
    kernel_size: typing.Optional[
        typing.Union[int, typing.Sequence[int]]
    ] = None,
    strides: int = 1,
    groups: int = 1,
    activation: typing.Optional[str] = None,
    use_depthwise: bool = False,
    add_skip: bool = False,
    bn_momentum: float = 0.9,
    bn_epsilon: float = 1e-5,
    padding: typing.Optional[typing.Literal["same", "valid"]] = None,
    name="conv2d_block",
):
    if kernel_size is None:
        raise ValueError(
            f"kernel_size must be passed. Received: kernel_size={kernel_size}"
        )
    if isinstance(kernel_size, int):
        kernel_size = [kernel_size, kernel_size]

    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    has_skip = add_skip and strides == 1 and input_channels == filters
    x = inputs

    if padding is None:
        padding = "same"
        if strides > 1:
            padding = "valid"
            x = layers.ZeroPadding2D(
                (kernel_size[0] // 2, kernel_size[1] // 2), name=f"{name}_pad"
            )(x)

    if not use_depthwise:
        x = layers.Conv2D(
            filters,
            kernel_size,
            strides,
            padding=padding,
            groups=groups,
            use_bias=False,
            name=f"{name}_conv2d",
        )(x)
    else:
        x = layers.DepthwiseConv2D(
            kernel_size,
            strides,
            padding=padding,
            use_bias=False,
            name=f"{name}_dwconv2d",
        )(x)
    x = layers.BatchNormalization(
        axis=channels_axis,
        name=f"{name}_bn",
        momentum=bn_momentum,
        epsilon=bn_epsilon,
    )(x)
    x = apply_activation(x, activation, name=name)
    if has_skip:
        x = layers.Add()([x, inputs])
    return x


def apply_se_block(
    inputs,
    se_ratio: float = 0.25,
    activation: typing.Optional[str] = "relu",
    gate_activation: typing.Optional[str] = "sigmoid",
    make_divisible_number: typing.Optional[int] = None,
    se_input_channels: typing.Optional[int] = None,
    name: str = "se_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    if se_input_channels is None:
        se_input_channels = input_channels
    if make_divisible_number is None:
        se_channels = round(se_input_channels * se_ratio)
    else:
        se_channels = make_divisible(
            se_input_channels * se_ratio, make_divisible_number
        )

    x = inputs
    x = layers.GlobalAveragePooling2D(
        data_format=backend.image_data_format(),
        keepdims=True,
        name=f"{name}_mean",
    )(x)
    x = layers.Conv2D(
        se_channels, 1, use_bias=True, name=f"{name}_conv_reduce"
    )(x)
    x = apply_activation(x, activation, name=f"{name}_act1")
    x = layers.Conv2D(
        input_channels, 1, use_bias=True, name=f"{name}_conv_expand"
    )(x)
    x = apply_activation(x, gate_activation, name=f"{name}_gate")
    x = layers.Multiply(name=name)([inputs, x])
    return x



import typing


def make_divisible(
    v,
    divisor: int = 8,
    min_value: typing.Optional[float] = None,
    round_limit: float = 0.9,
):
    min_value = min_value or divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < round_limit * v:
        new_v += divisor
    return new_v



import typing

import keras
from keras import backend
from keras import layers
from keras import ops

from kimm import layers as kimm_layers
from kimm.blocks import apply_transformer_block
from kimm.models.base_model import BaseModel
from kimm.utils import add_model_to_registry


@keras.saving.register_keras_serializable(package="kimm")
class VisionTransformer(BaseModel):
    def __init__(
        self,
        patch_size: int,
        embed_dim: int,
        depth: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        pos_dropout_rate: float = 0.0,
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs, 384)
        if self._pooling is not None:
            raise ValueError(
                "`VisionTransformer` doesn't support `pooling`. "
                f"Received: pooling={self._pooling}"
            )
        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
            static_shape=True,
        )
        x = inputs

        x = self.build_preprocessing(x, "-1_1")

        # Prepare feature extraction
        features = {}

        # Patch embedding
        x = layers.Conv2D(
            embed_dim,
            kernel_size=patch_size,
            strides=patch_size,
            padding="valid",
            use_bias=True,
            name="patch_embed_conv",
        )(x)

        # TODO: natively support channels_first
        if backend.image_data_format() == "channels_first":
            x = ops.transpose(x, [0, 2, 3, 1])

        x = layers.Reshape((-1, embed_dim))(x)
        x = kimm_layers.PositionEmbedding(name="postition_embedding")(x)
        features["EMBEDDING"] = x
        x = layers.Dropout(pos_dropout_rate, name="pos_dropout")(x)

        for i in range(depth):
            x = apply_transformer_block(
                x,
                embed_dim,
                num_heads,
                mlp_ratio,
                use_qkv_bias,
                use_qk_norm,
                activation="gelu",
                name=f"blocks_{i}",
            )
            features[f"BLOCK{i}"] = x
        x = layers.LayerNormalization(epsilon=1e-6, name="norm")(x)

        # Head
        if self._include_top:
            x = self.build_top(
                x,
                self._classes,
                self._classifier_activation,
                self._dropout_rate,
            )

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.patch_size = patch_size
        self.embed_dim = embed_dim
        self.depth = depth
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio
        self.use_qkv_bias = use_qkv_bias
        self.use_qk_norm = use_qk_norm
        self.pos_dropout_rate = pos_dropout_rate

    def build_top(self, inputs, classes, classifier_activation, dropout_rate):
        x = inputs[:, 0]  # class token
        x = layers.Dropout(dropout_rate, name="head_drop")(x)
        x = layers.Dense(
            classes, activation=classifier_activation, name="head"
        )(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "patch_size": self.patch_size,
                "embed_dim": self.embed_dim,
                "depth": self.depth,
                "num_heads": self.num_heads,
                "mlp_ratio": self.mlp_ratio,
                "use_qkv_bias": self.use_qkv_bias,
                "use_qk_norm": self.use_qk_norm,
                "pos_dropout_rate": self.pos_dropout_rate,
            }
        )
        return config

    def fix_config(self, config):
        unused_kwargs = [
            "patch_size",
            "embed_dim",
            "depth",
            "num_heads",
            "mlp_ratio",
            "use_qkv_bias",
            "use_qk_norm",
            "pos_dropout_rate",
        ]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class VisionTransformerTiny16(VisionTransformer):
    available_feature_keys = [
        "EMBEDDING",
        *[f"BLOCK{i}" for i in range(12)],
    ]
    available_weights = [
        (
            "imagenet",
            VisionTransformer.default_origin,
            "visiontransformertiny16_vit_tiny_patch16_384.keras",
        )
    ]

    def __init__(
        self,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        pos_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VisionTransformerTiny16",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            16,
            192,
            12,
            3,
            mlp_ratio,
            use_qkv_bias,
            use_qk_norm,
            pos_dropout_rate,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VisionTransformerTiny32(VisionTransformer):
    available_feature_keys = [
        "EMBEDDING",
        *[f"BLOCK{i}" for i in range(12)],
    ]
    available_weights = []

    def __init__(
        self,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        pos_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = None,
        name: str = "VisionTransformerTiny32",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            32,
            192,
            12,
            3,
            mlp_ratio,
            use_qkv_bias,
            use_qk_norm,
            pos_dropout_rate,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VisionTransformerSmall16(VisionTransformer):
    available_feature_keys = [
        "EMBEDDING",
        *[f"BLOCK{i}" for i in range(12)],
    ]
    available_weights = [
        (
            "imagenet",
            VisionTransformer.default_origin,
            "visiontransformersmall16_vit_small_patch16_384.keras",
        )
    ]

    def __init__(
        self,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        pos_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VisionTransformerSmall16",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            16,
            384,
            12,
            6,
            mlp_ratio,
            use_qkv_bias,
            use_qk_norm,
            pos_dropout_rate,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VisionTransformerSmall32(VisionTransformer):
    available_feature_keys = [
        "EMBEDDING",
        *[f"BLOCK{i}" for i in range(12)],
    ]
    available_weights = [
        (
            "imagenet",
            VisionTransformer.default_origin,
            "visiontransformersmall32_vit_small_patch32_384.keras",
        )
    ]

    def __init__(
        self,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        pos_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VisionTransformerSmall32",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            32,
            384,
            12,
            6,
            mlp_ratio,
            use_qkv_bias,
            use_qk_norm,
            pos_dropout_rate,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VisionTransformerBase16(VisionTransformer):
    available_feature_keys = [
        "EMBEDDING",
        *[f"BLOCK{i}" for i in range(12)],
    ]
    available_weights = [
        (
            "imagenet",
            VisionTransformer.default_origin,
            "visiontransformerbase16_vit_base_patch16_384.keras",
        )
    ]

    def __init__(
        self,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        pos_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VisionTransformerBase16",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            16,
            768,
            12,
            12,
            mlp_ratio,
            use_qkv_bias,
            use_qk_norm,
            pos_dropout_rate,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VisionTransformerBase32(VisionTransformer):
    available_feature_keys = [
        "EMBEDDING",
        *[f"BLOCK{i}" for i in range(12)],
    ]
    available_weights = [
        (
            "imagenet",
            VisionTransformer.default_origin,
            "visiontransformerbase32_vit_base_patch32_384.keras",
        )
    ]

    def __init__(
        self,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        pos_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VisionTransformerBase32",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            32,
            768,
            12,
            12,
            mlp_ratio,
            use_qkv_bias,
            use_qk_norm,
            pos_dropout_rate,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VisionTransformerLarge16(VisionTransformer):
    available_feature_keys = [
        "EMBEDDING",
        *[f"BLOCK{i}" for i in range(24)],
    ]
    available_weights = []

    def __init__(
        self,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        pos_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = None,
        name: str = "VisionTransformerLarge16",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            16,
            1024,
            24,
            16,
            mlp_ratio,
            use_qkv_bias,
            use_qk_norm,
            pos_dropout_rate,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VisionTransformerLarge32(VisionTransformer):
    available_feature_keys = [
        "EMBEDDING",
        *[f"BLOCK{i}" for i in range(24)],
    ]
    available_weights = []

    def __init__(
        self,
        mlp_ratio: float = 4.0,
        use_qkv_bias: bool = True,
        use_qk_norm: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        pos_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = None,
        name: str = "VisionTransformerLarge32",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            32,
            1024,
            24,
            16,
            mlp_ratio,
            use_qkv_bias,
            use_qk_norm,
            pos_dropout_rate,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(VisionTransformerTiny16, "imagenet")
add_model_to_registry(VisionTransformerTiny32)
add_model_to_registry(VisionTransformerSmall16, "imagenet")
add_model_to_registry(VisionTransformerSmall32, "imagenet")
add_model_to_registry(VisionTransformerBase16, "imagenet")
add_model_to_registry(VisionTransformerBase32, "imagenet")
add_model_to_registry(VisionTransformerLarge16)
add_model_to_registry(VisionTransformerLarge32)



import typing

from keras import backend
from keras import layers

from kimm.blocks.base_block import apply_conv2d_block
from kimm.blocks.base_block import apply_se_block
from kimm.utils import make_divisible


def apply_inverted_residual_block(
    inputs,
    output_channels: int,
    depthwise_kernel_size: int = 3,
    expansion_kernel_size: int = 1,
    pointwise_kernel_size: int = 1,
    strides: int = 1,
    expansion_ratio: float = 1.0,
    se_ratio: float = 0.0,
    activation: str = "swish",
    se_channels: typing.Optional[int] = None,
    se_activation: typing.Optional[str] = None,
    se_gate_activation: typing.Optional[str] = "sigmoid",
    se_make_divisible_number: typing.Optional[int] = None,
    bn_epsilon: float = 1e-5,
    padding: typing.Optional[typing.Literal["same", "valid"]] = None,
    name: str = "inverted_residual_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    hidden_channels = make_divisible(input_channels * expansion_ratio)
    has_skip = strides == 1 and input_channels == output_channels

    x = inputs
    # Point-wise expansion
    x = apply_conv2d_block(
        x,
        hidden_channels,
        expansion_kernel_size,
        1,
        activation=activation,
        bn_epsilon=bn_epsilon,
        padding=padding,
        name=f"{name}_conv_pw",
    )
    # Depth-wise convolution
    x = apply_conv2d_block(
        x,
        kernel_size=depthwise_kernel_size,
        strides=strides,
        activation=activation,
        use_depthwise=True,
        bn_epsilon=bn_epsilon,
        padding=padding,
        name=f"{name}_conv_dw",
    )
    # Squeeze-and-excitation
    if se_ratio > 0:
        x = apply_se_block(
            x,
            se_ratio,
            activation=se_activation or activation,
            gate_activation=se_gate_activation,
            se_input_channels=se_channels,
            make_divisible_number=se_make_divisible_number,
            name=f"{name}_se",
        )
    # Point-wise linear projection
    x = apply_conv2d_block(
        x,
        output_channels,
        pointwise_kernel_size,
        1,
        activation=None,
        bn_epsilon=bn_epsilon,
        padding=padding,
        name=f"{name}_conv_pwl",
    )
    if has_skip:
        x = layers.Add()([x, inputs])
    return x



import typing

import keras
from keras import backend
from keras import layers
from keras import ops

from kimm.blocks import apply_conv2d_block
from kimm.blocks import apply_se_block
from kimm.models import BaseModel
from kimm.utils import add_model_to_registry
from kimm.utils import make_divisible

DEFAULT_CONFIG = [
    # k, t, c, SE, s
    # stage1
    [
        [3, 16, 16, 0, 1],
    ],
    # stage2
    [
        [3, 48, 24, 0, 2],
    ],
    [
        [3, 72, 24, 0, 1],
    ],
    # stage3
    [
        [5, 72, 40, 0.25, 2],
    ],
    [
        [5, 120, 40, 0.25, 1],
    ],
    # stage4
    [
        [3, 240, 80, 0, 2],
    ],
    [
        [3, 200, 80, 0, 1],
        [3, 184, 80, 0, 1],
        [3, 184, 80, 0, 1],
        [3, 480, 112, 0.25, 1],
        [3, 672, 112, 0.25, 1],
    ],
    # stage5
    [
        [5, 672, 160, 0.25, 2],
    ],
    [
        [5, 960, 160, 0, 1],
        [5, 960, 160, 0.25, 1],
        [5, 960, 160, 0, 1],
        [5, 960, 160, 0.25, 1],
    ],
]


def apply_ghost_block(
    inputs,
    output_channels: int,
    expand_ratio: float = 2.0,
    kernel_size: int = 1,
    depthwise_kernel_size: int = 3,
    strides: int = 1,
    activation="relu",
    name="ghost_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    hidden_channels_1 = int(ops.ceil(output_channels / expand_ratio))
    hidden_channels_2 = int(hidden_channels_1 * (expand_ratio - 1.0))

    x = inputs
    x1 = apply_conv2d_block(
        x,
        hidden_channels_1,
        kernel_size,
        strides,
        activation=activation,
        name=f"{name}_primary_conv",
    )
    x2 = apply_conv2d_block(
        x1,
        hidden_channels_2,
        depthwise_kernel_size,
        1,
        activation=activation,
        use_depthwise=True,
        name=f"{name}_cheap_operation",
    )
    out = layers.Concatenate(axis=channels_axis, name=f"{name}")([x1, x2])
    if channels_axis == -1:
        return out[..., :output_channels]
    else:
        return out[:, :output_channels, ...]


def apply_ghost_block_v2(
    inputs,
    output_channels: int,
    expand_ratio: float = 2.0,
    kernel_size: int = 1,
    depthwise_kernel_size: int = 3,
    strides: int = 1,
    activation="relu",
    name="ghost_block_v2",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    if backend.image_data_format() == "channels_last":
        output_axis = (-3, -2)
    else:
        output_axis = (-2, -1)

    hidden_channels_1 = int(ops.ceil(output_channels / expand_ratio))
    hidden_channels_2 = int(hidden_channels_1 * (expand_ratio - 1.0))

    x = inputs
    residual = inputs
    x1 = apply_conv2d_block(
        x,
        hidden_channels_1,
        kernel_size,
        strides,
        activation=activation,
        name=f"{name}_primary_conv",
    )
    x2 = apply_conv2d_block(
        x1,
        hidden_channels_2,
        depthwise_kernel_size,
        1,
        activation=activation,
        use_depthwise=True,
        name=f"{name}_cheap_operation",
    )
    out = layers.Concatenate(axis=channels_axis, name=f"{name}_concat")(
        [x1, x2]
    )

    residual = layers.AveragePooling2D(
        2, 2, data_format=backend.image_data_format(), name=f"{name}_avg_pool"
    )(residual)
    residual = apply_conv2d_block(
        residual,
        output_channels,
        kernel_size,
        strides,
        name=f"{name}_short_conv1",
    )
    residual = apply_conv2d_block(
        residual,
        output_channels,
        (1, 5),
        1,
        use_depthwise=True,
        name=f"{name}_short_conv2",
    )
    residual = apply_conv2d_block(
        residual,
        output_channels,
        (5, 1),
        1,
        use_depthwise=True,
        name=f"{name}_short_conv3",
    )
    residual = layers.Activation("sigmoid", name=f"{name}_gate")(residual)
    # TODO: support dynamic shape
    residual = layers.Resizing(
        out.shape[output_axis[0]], out.shape[output_axis[1]], "nearest"
    )(residual)
    if channels_axis == -1:
        out = out[..., :output_channels]
    else:
        out = out[:, :output_channels, ...]
    out = layers.Multiply(name=name)([out, residual])
    return out


def apply_ghost_bottleneck(
    inputs,
    hidden_channels: int,
    output_channels: int,
    depthwise_kernel_size: int = 3,
    strides: int = 1,
    se_ratio: float = 0.0,
    activation="relu",
    use_attention=False,  # GhostNetV2
    name="ghost_bottlenect",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    has_se = se_ratio is not None and se_ratio > 0.0

    x = inputs
    shortcut = inputs
    # point-wise
    if use_attention:
        x = apply_ghost_block_v2(
            x, hidden_channels, activation=activation, name=f"{name}_ghost1"
        )
    else:
        x = apply_ghost_block(
            x, hidden_channels, activation=activation, name=f"{name}_ghost1"
        )
    # depthwise
    if strides > 1:
        x = apply_conv2d_block(
            x,
            hidden_channels,
            depthwise_kernel_size,
            strides,
            use_depthwise=True,
            name=f"{name}_conv_dw",
        )
    # squeeze-and-excitation
    if has_se:
        x = apply_se_block(
            x,
            se_ratio,
            gate_activation="hard_sigmoid",
            make_divisible_number=4,
            name=f"{name}_se",
        )
    # point-wise
    x = apply_ghost_block(
        x, output_channels, activation=None, name=f"{name}_ghost2"
    )
    # shortcut
    if input_channels != output_channels or strides > 1:
        shortcut = apply_conv2d_block(
            shortcut,
            input_channels,
            depthwise_kernel_size,
            strides,
            activation=None,
            use_depthwise=True,
            name=f"{name}_shortcut1",
        )
        shortcut = apply_conv2d_block(
            shortcut,
            output_channels,
            1,
            1,
            activation=None,
            name=f"{name}_shortcut2",
        )
    out = layers.Add(name=name)([x, shortcut])
    return out


@keras.saving.register_keras_serializable(package="kimm")
class GhostNet(BaseModel):
    available_feature_keys = [
        "STEM_S2",
        *[
            f"BLOCK{i}_S{j}"
            for i, j in zip(range(9), [2, 4, 4, 8, 8, 16, 16, 32, 32])
        ],
    ]

    def __init__(
        self,
        width: float = 1.0,
        config: typing.Union[str, typing.List] = "default",
        version: typing.Literal["v1", "v2"] = "v1",
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        _available_configs = ["default"]
        if config == "default":
            _config = DEFAULT_CONFIG
        else:
            raise ValueError(
                f"config must be one of {_available_configs} using string. "
                f"Received: config={config}"
            )
        if version not in ("v1", "v2"):
            raise ValueError(
                "`version` must be one of ('v1', 'v2'). "
                f"Received version={version}"
            )

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
            require_flatten=self._include_top,
            static_shape=True if version == "v2" else False,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # stem
        stem_channels = make_divisible(16 * width, 4)
        x = apply_conv2d_block(
            x, stem_channels, 3, 2, activation="relu", name="conv_stem"
        )
        features["STEM_S2"] = x

        # blocks
        total_layer_idx = 0
        current_stride = 2
        for current_block_idx, cfg in enumerate(_config):
            for current_layer_idx, (k, e, c, se, s) in enumerate(cfg):
                output_channels = make_divisible(c * width, 4)
                hidden_channels = make_divisible(e * width, 4)
                use_attention = False
                if version == "v2" and total_layer_idx > 1:
                    use_attention = True
                name = f"blocks_{current_block_idx}_{current_layer_idx}"
                x = apply_ghost_bottleneck(
                    x,
                    hidden_channels,
                    output_channels,
                    k,
                    s,
                    se_ratio=se,
                    use_attention=use_attention,
                    name=name,
                )
                total_layer_idx += 1
            current_stride *= s
            features[f"BLOCK{current_block_idx}_S{current_stride}"] = x
        # post stages conv block
        output_channels = make_divisible(e * width, 4)
        x = apply_conv2d_block(
            x,
            output_channels,
            1,
            activation="relu",
            name=f"blocks_{current_block_idx+1}",
        )

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.width = width
        self.config = config
        self.version = version

    def build_top(self, inputs, classes, classifier_activation, dropout_rate):
        x = layers.GlobalAveragePooling2D(name="avg_pool", keepdims=True)(
            inputs
        )
        x = layers.Conv2D(
            1280, 1, 1, use_bias=True, activation="relu", name="conv_head"
        )(x)
        x = layers.Flatten()(x)
        x = layers.Dropout(rate=dropout_rate, name="conv_head_dropout")(x)
        x = layers.Dense(
            classes, activation=classifier_activation, name="classifier"
        )(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "width": self.width,
                "config": self.config,
                "version": self.version,
            }
        )
        return config

    def fix_config(self, config):
        unused_kwargs = ["width", "config", "version"]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class GhostNet050(GhostNet):
    available_weights = []

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.2,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = None,
        config: typing.Union[str, typing.List] = "default",
        name: str = "GhostNet050",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            0.5,
            config,
            "v1",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class GhostNet100(GhostNet):
    available_weights = [
        (
            "imagenet",
            GhostNet.default_origin,
            "ghostnet100_ghostnet_100.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.2,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "GhostNet100",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.0,
            config,
            "v1",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class GhostNet130(GhostNet):
    available_weights = []

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.2,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = None,
        config: typing.Union[str, typing.List] = "default",
        name: str = "GhostNet130",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.3,
            config,
            "v1",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class GhostNet100V2(GhostNet):
    available_weights = [
        (
            "imagenet",
            GhostNet.default_origin,
            "ghostnet100v2_ghostnetv2_100.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.2,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "GhostNet100V2",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.0,
            config,
            "v2",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class GhostNet130V2(GhostNet):
    available_weights = [
        (
            "imagenet",
            GhostNet.default_origin,
            "ghostnet130v2_ghostnetv2_130.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.2,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "GhostNet130V2",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.3,
            config,
            "v2",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class GhostNet160V2(GhostNet):
    available_weights = [
        (
            "imagenet",
            GhostNet.default_origin,
            "ghostnet160v2_ghostnetv2_160.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.2,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "GhostNet160V2",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.6,
            config,
            "v2",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(GhostNet050)
add_model_to_registry(GhostNet100, "imagenet")
add_model_to_registry(GhostNet130)
add_model_to_registry(GhostNet100V2, "imagenet")
add_model_to_registry(GhostNet130V2, "imagenet")
add_model_to_registry(GhostNet160V2, "imagenet")



import math
import typing

import keras

from kimm.blocks import apply_conv2d_block
from kimm.blocks import apply_depthwise_separation_block
from kimm.blocks import apply_inverted_residual_block
from kimm.models.base_model import BaseModel
from kimm.utils import add_model_to_registry
from kimm.utils import make_divisible

DEFAULT_CONFIG = [
    # type, repeat, kernel_size, strides, expansion_ratio, channels
    ["ds", 1, 3, 1, 1, 16],
    ["ir", 2, 3, 2, 6, 24],
    ["ir", 3, 3, 2, 6, 32],
    ["ir", 4, 3, 2, 6, 64],
    ["ir", 3, 3, 1, 6, 96],
    ["ir", 3, 3, 2, 6, 160],
    ["ir", 1, 3, 1, 6, 320],
]


@keras.saving.register_keras_serializable(package="kimm")
class MobileNetV2(BaseModel):
    available_feature_keys = [
        "STEM_S2",
        *[
            f"BLOCK{i}_S{j}"
            for i, j in zip(range(7), [2, 4, 8, 16, 16, 32, 32])
        ],
    ]

    def __init__(
        self,
        width: float = 1.0,
        depth: float = 1.0,
        fix_stem_and_head_channels: bool = False,
        config: typing.Literal["default"] = "default",
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        _available_configs = ["default"]
        if config == "default":
            _config = DEFAULT_CONFIG
        else:
            raise ValueError(
                f"config must be one of {_available_configs} using string. "
                f"Received: config={config}"
            )

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # stem
        stem_channel = (
            32 if fix_stem_and_head_channels else make_divisible(32 * width)
        )
        x = apply_conv2d_block(
            x,
            stem_channel,
            3,
            2,
            activation="relu6",
            name="conv_stem",
        )
        features["STEM_S2"] = x

        # blocks
        current_stride = 2
        for current_block_idx, cfg in enumerate(_config):
            block_type, r, k, s, e, c = cfg
            c = make_divisible(c * width)
            # no depth multiplier at first and last block
            if current_block_idx not in (0, len(_config) - 1):
                r = int(math.ceil(r * depth))
            for current_layer_idx in range(r):
                s = s if current_layer_idx == 0 else 1
                name = f"blocks_{current_block_idx}_{current_layer_idx}"
                if block_type == "ds":
                    x = apply_depthwise_separation_block(
                        x, c, k, 1, s, activation="relu6", name=name
                    )
                elif block_type == "ir":
                    x = apply_inverted_residual_block(
                        x, c, k, 1, 1, s, e, activation="relu6", name=name
                    )
                current_stride *= s
            features[f"BLOCK{current_block_idx}_S{current_stride}"] = x

        # last conv
        if fix_stem_and_head_channels:
            head_channels = 1280
        else:
            head_channels = max(1280, make_divisible(1280 * width))
        x = apply_conv2d_block(
            x, head_channels, 1, 1, activation="relu6", name="conv_head"
        )

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.width = width
        self.depth = depth
        self.fix_stem_and_head_channels = fix_stem_and_head_channels
        self.config = config

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "width": self.width,
                "depth": self.depth,
                "fix_stem_and_head_channels": self.fix_stem_and_head_channels,
                "config": self.config,
            }
        )
        return config

    def fix_config(self, config):
        unused_kwargs = [
            "width",
            "depth",
            "fix_stem_and_head_channels",
            "config",
        ]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class MobileNet050V2(MobileNetV2):
    available_weights = [
        (
            "imagenet",
            MobileNetV2.default_origin,
            "mobilenet050v2_mobilenetv2_050.lamb_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "MobileNet050V2",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            0.5,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileNet100V2(MobileNetV2):
    available_weights = [
        (
            "imagenet",
            MobileNetV2.default_origin,
            "mobilenet100v2_mobilenetv2_100.ra_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "MobileNet100V2",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.0,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileNet110V2(MobileNetV2):
    available_weights = [
        (
            "imagenet",
            MobileNetV2.default_origin,
            "mobilenet110v2_mobilenetv2_110d.ra_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "MobileNet110V2",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.1,
            1.2,
            True,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileNet120V2(MobileNetV2):
    available_weights = [
        (
            "imagenet",
            MobileNetV2.default_origin,
            "mobilenet120v2_mobilenetv2_120d.ra_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "MobileNet120V2",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.2,
            1.4,
            True,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileNet140V2(MobileNetV2):
    available_weights = [
        (
            "imagenet",
            MobileNetV2.default_origin,
            "mobilenet140v2_mobilenetv2_140.ra_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "default",
        name: str = "MobileNet140V2",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.4,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(MobileNet050V2, "imagenet")
add_model_to_registry(MobileNet100V2, "imagenet")
add_model_to_registry(MobileNet110V2, "imagenet")
add_model_to_registry(MobileNet120V2, "imagenet")
add_model_to_registry(MobileNet140V2, "imagenet")



import typing

import keras
from keras import backend
from keras import layers

from kimm.models import BaseModel
from kimm.utils import add_model_to_registry


def apply_xception_block(
    inputs,
    depth: int,
    output_channels: int,
    strides: int = 1,
    no_first_act=False,
    grow_first=True,
    name="xception_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    x = inputs
    residual = inputs

    current_layer_idx = 0
    for i in range(depth):
        if grow_first:
            c = output_channels
        else:
            c = input_channels if i < (depth - 1) else output_channels
        if not no_first_act or i > 0:
            x = layers.ReLU(name=f"{name}_rep_{current_layer_idx}")(x)
            current_layer_idx += 1
        x = layers.SeparableConv2D(
            c,
            3,
            1,
            "same",
            use_bias=False,
            name=f"{name}_rep_{current_layer_idx}",
        )(x)
        x = layers.BatchNormalization(
            axis=channels_axis,
            name=f"{name}_rep_{current_layer_idx + 1}",
        )(x)
        current_layer_idx += 2

    if strides != 1:
        x = layers.MaxPooling2D(
            3, strides, padding="same", name=f"{name}_rep_{current_layer_idx}"
        )(x)
        current_layer_idx += 1

    if output_channels != input_channels or strides != 1:
        residual = layers.Conv2D(
            output_channels,
            1,
            strides,
            use_bias=False,
            name=f"{name}_skipconv2d",
        )(residual)
        residual = layers.BatchNormalization(
            axis=channels_axis, name=f"{name}_skipbn"
        )(residual)

    x = layers.Add()([x, residual])
    return x


@keras.saving.register_keras_serializable(package="kimm")
class XceptionBase(BaseModel):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(4), [4, 8, 16, 32])],
    ]

    def __init__(self, **kwargs):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        channels_axis = (
            -1 if backend.image_data_format() == "channels_last" else -3
        )

        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
            min_size=71,
        )
        x = inputs

        x = self.build_preprocessing(x, "-1_1")

        # Prepare feature extraction
        features = {}

        # Stem
        x = layers.Conv2D(32, 3, 2, use_bias=False, name="conv1")(x)
        x = layers.BatchNormalization(axis=channels_axis, name="bn1")(x)
        x = layers.ReLU()(x)
        x = layers.Conv2D(64, 3, 1, use_bias=False, name="conv2")(x)
        x = layers.BatchNormalization(axis=channels_axis, name="bn2")(x)
        x = layers.ReLU()(x)
        features["STEM_S2"] = x

        # Blocks
        x = apply_xception_block(x, 2, 128, 2, no_first_act=True, name="block1")
        features["BLOCK0_S4"] = x
        x = apply_xception_block(x, 2, 256, 2, name="block2")
        features["BLOCK1_S8"] = x
        x = apply_xception_block(x, 2, 728, 2, name="block3")

        x = apply_xception_block(x, 3, 728, 1, name="block4")
        x = apply_xception_block(x, 3, 728, 1, name="block5")
        x = apply_xception_block(x, 3, 728, 1, name="block6")
        x = apply_xception_block(x, 3, 728, 1, name="block7")

        x = apply_xception_block(x, 3, 728, 1, name="block8")
        x = apply_xception_block(x, 3, 728, 1, name="block9")
        x = apply_xception_block(x, 3, 728, 1, name="block10")
        x = apply_xception_block(x, 3, 728, 1, name="block11")
        features["BLOCK2_S16"] = x

        x = apply_xception_block(
            x, 2, 1024, 2, grow_first=False, name="block12"
        )

        x = layers.SeparableConv2D(
            1536, 3, 1, "same", use_bias=False, name="conv3"
        )(x)
        x = layers.BatchNormalization(axis=channels_axis, name="bn3")(x)
        x = layers.ReLU()(x)

        x = layers.SeparableConv2D(
            2048, 3, 1, "same", use_bias=False, name="conv4"
        )(x)
        x = layers.BatchNormalization(axis=channels_axis, name="bn4")(x)
        x = layers.ReLU()(x)
        features["BLOCK3_S32"] = x

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line

    def get_config(self):
        return super().get_config()

    def fix_config(self, config: typing.Dict):
        return config


"""
Model Definition
"""


class Xception(XceptionBase):
    available_weights = [
        (
            "imagenet",
            XceptionBase.default_origin,
            "xception.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "Xception",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            default_size=299,
            **kwargs,
        )


add_model_to_registry(Xception, "imagenet")



import keras
from keras import initializers
from keras import layers
from keras import ops
from keras.initializers import Initializer


@keras.saving.register_keras_serializable(package="kimm")
class LayerScale(layers.Layer):
    def __init__(
        self,
        axis: int = -1,
        initializer: Initializer = initializers.Constant(1e-5),
        name: str = "layer_scale",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.axis = axis
        self.initializer = initializer
        self.name = name

    def build(self, input_shape):
        if isinstance(self.axis, list):
            shape = tuple([input_shape[dim] for dim in self.axis])
        else:
            shape = (input_shape[self.axis],)
            self.axis = [self.axis]
        self.gamma = self.add_weight(
            shape, initializer=self.initializer, name="gamma"
        )
        self.built = True

    def call(self, inputs, training=None, mask=None):
        inputs = ops.cast(inputs, self.compute_dtype)

        # Broadcasting only necessary for norm when the axis is not just
        # the last dimension
        input_shape = inputs.shape
        ndims = len(inputs.shape)
        broadcast_shape = [1] * ndims
        for dim in self.axis:
            broadcast_shape[dim] = input_shape[dim]
        gamma = ops.reshape(self.gamma, broadcast_shape)
        gamma = ops.cast(gamma, self.compute_dtype)

        return ops.multiply(inputs, gamma)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "axis": self.axis,
                "initializer": initializers.serialize(self.initializer),
                "name": self.name,
            }
        )
        return config



import functools
import typing

import keras
from keras import backend
from keras import layers

from kimm.blocks import apply_conv2d_block
from kimm.models import BaseModel
from kimm.utils import add_model_to_registry

_apply_conv2d_block = functools.partial(
    apply_conv2d_block, activation="relu", bn_epsilon=1e-3, padding="valid"
)


def apply_inception_a_block(inputs, pool_channels, name="inception_a_block"):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3

    x = inputs

    branch1x1 = _apply_conv2d_block(x, 64, 1, 1, name=f"{name}_branch1x1")

    branch5x5 = _apply_conv2d_block(x, 48, 1, 1, name=f"{name}_branch5x5_1")
    branch5x5 = _apply_conv2d_block(
        branch5x5, 64, 5, 1, padding=None, name=f"{name}_branch5x5_2"
    )

    branch3x3dbl = _apply_conv2d_block(
        x, 64, 1, 1, name=f"{name}_branch3x3dbl_1"
    )
    branch3x3dbl = _apply_conv2d_block(
        branch3x3dbl, 96, 3, 1, padding=None, name=f"{name}_branch3x3dbl_2"
    )
    branch3x3dbl = _apply_conv2d_block(
        branch3x3dbl, 96, 3, 1, padding=None, name=f"{name}_branch3x3dbl_3"
    )

    branch_pool = layers.ZeroPadding2D(1)(x)
    branch_pool = layers.AveragePooling2D(
        3, 1, data_format=backend.image_data_format()
    )(branch_pool)
    branch_pool = _apply_conv2d_block(
        branch_pool,
        pool_channels,
        1,
        1,
        activation="relu",
        name=f"{name}_branch_pool",
    )
    x = layers.Concatenate(axis=channels_axis)(
        [branch1x1, branch5x5, branch3x3dbl, branch_pool]
    )
    return x


def apply_inception_b_block(inputs, name="incpetion_b_block"):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3

    x = inputs

    branch3x3 = _apply_conv2d_block(x, 384, 3, 2, name=f"{name}_branch3x3")

    branch3x3dbl = _apply_conv2d_block(
        x, 64, 1, 1, name=f"{name}_branch3x3dbl_1"
    )
    branch3x3dbl = _apply_conv2d_block(
        branch3x3dbl, 96, 3, 1, padding=None, name=f"{name}_branch3x3dbl_2"
    )
    branch3x3dbl = _apply_conv2d_block(
        branch3x3dbl, 96, 3, 2, name=f"{name}_branch3x3dbl_3"
    )

    branch_pool = layers.MaxPooling2D(3, 2, name=f"{name}_branch_pool")(x)
    x = layers.Concatenate(axis=channels_axis)(
        [branch3x3, branch3x3dbl, branch_pool]
    )
    return x


def apply_inception_c_block(
    inputs, branch7x7_channels, name="inception_c_block"
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    c7 = branch7x7_channels

    x = inputs

    branch1x1 = _apply_conv2d_block(x, 192, 1, 1, name=f"{name}_branch1x1")

    branch7x7 = _apply_conv2d_block(x, c7, 1, 1, name=f"{name}_branch7x7_1")
    branch7x7 = _apply_conv2d_block(
        branch7x7, c7, (1, 7), 1, padding=None, name=f"{name}_branch7x7_2"
    )
    branch7x7 = _apply_conv2d_block(
        branch7x7, 192, (7, 1), 1, padding=None, name=f"{name}_branch7x7_3"
    )

    branch7x7dbl = _apply_conv2d_block(
        x, c7, 1, 1, name=f"{name}_branch7x7dbl_1"
    )
    branch7x7dbl = _apply_conv2d_block(
        branch7x7dbl, c7, (7, 1), 1, padding=None, name=f"{name}_branch7x7dbl_2"
    )
    branch7x7dbl = _apply_conv2d_block(
        branch7x7dbl, c7, (1, 7), 1, padding=None, name=f"{name}_branch7x7dbl_3"
    )
    branch7x7dbl = _apply_conv2d_block(
        branch7x7dbl, c7, (7, 1), 1, padding=None, name=f"{name}_branch7x7dbl_4"
    )
    branch7x7dbl = _apply_conv2d_block(
        branch7x7dbl,
        192,
        (1, 7),
        1,
        padding=None,
        name=f"{name}_branch7x7dbl_5",
    )

    branch_pool = layers.ZeroPadding2D(1)(x)
    branch_pool = layers.AveragePooling2D(
        3, 1, data_format=backend.image_data_format()
    )(branch_pool)
    branch_pool = _apply_conv2d_block(
        branch_pool, 192, 1, 1, name=f"{name}_branch_pool"
    )
    x = layers.Concatenate(axis=channels_axis)(
        [branch1x1, branch7x7, branch7x7dbl, branch_pool]
    )
    return x


def apply_inception_d_block(inputs, name="inception_d_block"):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3

    x = inputs

    branch3x3 = _apply_conv2d_block(x, 192, 1, 1, name=f"{name}_branch3x3_1")
    branch3x3 = _apply_conv2d_block(
        branch3x3, 320, 3, 2, name=f"{name}_branch3x3_2"
    )

    branch7x7x3 = _apply_conv2d_block(
        x, 192, 1, 1, name=f"{name}_branch7x7x3_1"
    )
    branch7x7x3 = _apply_conv2d_block(
        branch7x7x3, 192, (1, 7), 1, padding=None, name=f"{name}_branch7x7x3_2"
    )
    branch7x7x3 = _apply_conv2d_block(
        branch7x7x3, 192, (7, 1), 1, padding=None, name=f"{name}_branch7x7x3_3"
    )
    branch7x7x3 = _apply_conv2d_block(
        branch7x7x3, 192, 3, 2, name=f"{name}_branch7x7x3_4"
    )

    branch_pool = layers.MaxPooling2D(3, 2)(x)
    x = layers.Concatenate(axis=channels_axis)(
        [branch3x3, branch7x7x3, branch_pool]
    )
    return x


def apply_inception_e_block(inputs, name="inception_e_block"):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3

    x = inputs

    branch1x1 = _apply_conv2d_block(x, 320, 1, 1, name=f"{name}_branch1x1")

    branch3x3 = _apply_conv2d_block(x, 384, 1, 1, name=f"{name}_branch3x3_1")
    branch3x3 = [
        _apply_conv2d_block(
            branch3x3, 384, (1, 3), 1, padding=None, name=f"{name}_branch3x3_2a"
        ),
        _apply_conv2d_block(
            branch3x3, 384, (3, 1), 1, padding=None, name=f"{name}_branch3x3_2b"
        ),
    ]
    branch3x3 = layers.Concatenate(axis=channels_axis)(branch3x3)

    branch3x3dbl = _apply_conv2d_block(
        x, 448, 1, 1, name=f"{name}_branch3x3dbl_1"
    )
    branch3x3dbl = _apply_conv2d_block(
        branch3x3dbl, 384, 3, 1, padding=None, name=f"{name}_branch3x3dbl_2"
    )
    branch3x3dbl = [
        _apply_conv2d_block(
            branch3x3dbl,
            384,
            (1, 3),
            1,
            padding=None,
            name=f"{name}_branch3x3dbl_3a",
        ),
        _apply_conv2d_block(
            branch3x3dbl,
            384,
            (3, 1),
            1,
            padding=None,
            name=f"{name}_branch3x3dbl_3b",
        ),
    ]
    branch3x3dbl = layers.Concatenate(axis=channels_axis)(branch3x3dbl)

    branch_pool = layers.ZeroPadding2D(1)(x)
    branch_pool = layers.AveragePooling2D(
        3, 1, data_format=backend.image_data_format()
    )(branch_pool)
    branch_pool = _apply_conv2d_block(
        branch_pool, 192, 1, 1, name=f"{name}_branch_pool"
    )
    x = layers.Concatenate(axis=channels_axis)(
        [branch1x1, branch3x3, branch3x3dbl, branch_pool]
    )
    return x


def apply_inception_aux_block(inputs, classes, name="inception_aux_block"):
    x = inputs

    x = layers.AveragePooling2D(5, 3, data_format=backend.image_data_format())(
        x
    )
    x = _apply_conv2d_block(x, 128, 1, 1, name=f"{name}_conv0")
    x = _apply_conv2d_block(x, 768, 5, 1, name=f"{name}_conv1")
    x = layers.GlobalAveragePooling2D()(x)
    x = layers.Dense(classes, use_bias=True, name=f"{name}_fc")(x)
    return x


@keras.saving.register_keras_serializable(package="kimm")
class InceptionV3Base(BaseModel):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(4), [4, 8, 16, 32])],
    ]

    def __init__(self, has_aux_logits: bool = False, **kwargs):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs, 299)
        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
            require_flatten=self._include_top,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # Stem block
        x = _apply_conv2d_block(x, 32, 3, 2, name="Conv2d_1a_3x3")
        x = _apply_conv2d_block(x, 32, 3, 1, name="Conv2d_2a_3x3")
        x = _apply_conv2d_block(x, 64, 3, 1, padding=None, name="Conv2d_2b_3x3")
        features["STEM_S2"] = x

        # Blocks
        x = layers.MaxPooling2D(3, 2, name="Pool1")(x)
        x = _apply_conv2d_block(x, 80, 1, 1, name="Conv2d_3b_1x1")
        x = _apply_conv2d_block(x, 192, 3, 1, name="Conv2d_4a_3x3")
        features["BLOCK0_S4"] = x
        x = layers.MaxPooling2D(3, 2, name="Pool2")(x)
        x = apply_inception_a_block(x, 32, "Mixed_5b")
        x = apply_inception_a_block(x, 64, "Mixed_5c")
        x = apply_inception_a_block(x, 64, "Mixed_5d")
        features["BLOCK1_S8"] = x

        x = apply_inception_b_block(x, "Mixed_6a")

        x = apply_inception_c_block(x, 128, "Mixed_6b")
        x = apply_inception_c_block(x, 160, "Mixed_6c")
        x = apply_inception_c_block(x, 160, "Mixed_6d")
        x = apply_inception_c_block(x, 192, "Mixed_6e")
        features["BLOCK2_S16"] = x

        if has_aux_logits:
            aux_logits = apply_inception_aux_block(
                x, self._classes, "AuxLogits"
            )

        x = apply_inception_d_block(x, "Mixed_7a")
        x = apply_inception_e_block(x, "Mixed_7b")
        x = apply_inception_e_block(x, "Mixed_7c")
        features["BLOCK3_S32"] = x

        # Head
        x = self.build_head(x)

        if has_aux_logits:
            x = [x, aux_logits]
        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.has_aux_logits = has_aux_logits

    def get_config(self):
        config = super().get_config()
        config.update({"has_aux_logits": self.has_aux_logits})
        return config

    def fix_config(self, config: typing.Dict):
        return config


class InceptionV3(InceptionV3Base):
    available_weights = [
        (
            "imagenet_aux_logits",
            InceptionV3Base.default_origin,
            "inceptionv3_inception_v3.gluon_in1k_aux_logits.keras",
        ),
        (
            "imagenet_no_aux_logits",
            InceptionV3Base.default_origin,
            "inceptionv3_inception_v3.gluon_in1k_no_aux_logits.keras",
        ),
    ]

    def __init__(
        self,
        has_aux_logits: bool = False,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "InceptionV3",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        if weights == "imagenet":
            if has_aux_logits:
                weights = f"{weights}_aux_logits"
            else:
                weights = f"{weights}_no_aux_logits"
        super().__init__(
            has_aux_logits,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(InceptionV3, "imagenet")



import sys
import typing
import warnings

# {
#     "name",  # str
#     "feature_extractor",  # bool
#     "feature_keys",  # list of str
#     "weights",  # None or str
# }
MODEL_REGISTRY: typing.List[typing.Dict[str, typing.Union[str, bool]]] = []


def _match_string(query: str, target: str):
    query = query.lower().replace(" ", "").replace("_", "").replace(".", "")
    target = target.lower()
    matched_idx = -1
    for q_char in query:
        matched = False
        for idx, t_char in enumerate(target):
            if matched:
                break
            if q_char == t_char and idx > matched_idx:
                matched_idx = idx
                matched = True
        if not matched:
            return False
    return True


def clear_registry():
    MODEL_REGISTRY.clear()


def add_model_to_registry(model_cls, weights: typing.Optional[str] = None):
    from kimm.models.base_model import BaseModel

    # deal with __all__
    mod = sys.modules[model_cls.__module__]
    model_name = model_cls.__name__
    if hasattr(mod, "__all__"):
        mod.__all__.append(model_name)
    else:
        mod.__all__ = [model_name]

    # add model information
    feature_extractor = False
    feature_keys = []
    if issubclass(model_cls, BaseModel):
        feature_extractor = True
        feature_keys = model_cls.available_feature_keys
    for info in MODEL_REGISTRY:
        if info["name"] == model_cls.__name__:
            warnings.warn(
                f"MODEL_REGISTRY already contains name={model_cls.__name__}!"
            )
    if weights is not None:
        if not isinstance(weights, str):
            raise ValueError(
                "`weights` must be one of (None, str). "
                f"Recieved: weight={weights}"
            )
        weights = weights.lower()
    MODEL_REGISTRY.append(
        {
            "name": model_cls.__name__,
            "feature_extractor": feature_extractor,
            "feature_keys": feature_keys,
            "weights": weights,
        }
    )


def list_models(
    name: typing.Optional[str] = None,
    feature_extractor: typing.Optional[bool] = None,
    weights: typing.Optional[typing.Union[bool, str]] = None,
):
    result_names: typing.Set = set()
    for info in MODEL_REGISTRY:
        # add by default
        result_names.add(info["name"])
        need_remove = False

        # match string (simple implementation)
        if name is not None:
            need_remove = not _match_string(name, info["name"])

        # filter by feature_extractor and weights
        if (
            feature_extractor is not None
            and info["feature_extractor"] is not feature_extractor
        ):
            need_remove = True
        if weights is not None and info["weights"] != weights:
            if weights is True and info["weights"] is None:
                need_remove = True
            elif weights is False and info["weights"] is not None:
                need_remove = True
            elif isinstance(weights, str):
                if weights.lower() != info["weights"]:
                    need_remove = True

        if need_remove:
            result_names.remove(info["name"])
    return sorted(result_names)



import typing

import keras
from keras import backend
from keras import layers

from kimm.blocks import apply_conv2d_block
from kimm.models.base_model import BaseModel
from kimm.utils import add_model_to_registry


def apply_basic_block(
    inputs,
    output_channels: int,
    strides: int = 1,
    activation="relu",
    name="basic_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    shortcut = inputs
    x = inputs
    x = apply_conv2d_block(
        x,
        output_channels,
        3,
        strides,
        activation=activation,
        name=f"{name}_conv1",
    )
    x = apply_conv2d_block(
        x,
        output_channels,
        3,
        1,
        activation=None,
        name=f"{name}_conv2",
    )

    # downsampling
    if strides != 1 or input_channels != output_channels:
        shortcut = apply_conv2d_block(
            shortcut,
            output_channels,
            1,
            strides,
            activation=None,
            name=f"{name}_downsample",
        )

    x = layers.Add(name=f"{name}_add")([x, shortcut])
    x = layers.Activation(activation=activation, name=f"{name}")(x)
    return x


def apply_bottleneck_block(
    inputs,
    output_channels: int,
    strides: int = 1,
    activation="relu",
    name="bottleneck_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    expansion = 4
    shortcut = inputs
    x = inputs
    x = apply_conv2d_block(
        x,
        output_channels,
        1,
        1,
        activation=activation,
        name=f"{name}_conv1",
    )
    x = apply_conv2d_block(
        x,
        output_channels,
        3,
        strides,
        activation=activation,
        name=f"{name}_conv2",
    )
    x = apply_conv2d_block(
        x,
        output_channels * expansion,
        1,
        1,
        activation=None,
        name=f"{name}_conv3",
    )

    # downsampling
    if strides != 1 or input_channels != output_channels * expansion:
        shortcut = apply_conv2d_block(
            shortcut,
            output_channels * expansion,
            1,
            strides,
            activation=None,
            name=f"{name}_downsample",
        )

    x = layers.Add(name=f"{name}_add")([x, shortcut])
    x = layers.Activation(activation=activation, name=f"{name}")(x)
    return x


@keras.saving.register_keras_serializable(package="kimm")
class ResNet(BaseModel):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(4), [4, 8, 16, 32])],
    ]

    def __init__(
        self,
        block_fn: typing.Literal["basic", "bottleneck"],
        num_blocks: typing.Sequence[int],
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        if block_fn not in ("basic", "bottleneck"):
            raise ValueError(
                "`block_fn` must be one of ('basic', 'bottelneck'). "
                f"Received: block_fn={block_fn}"
            )

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # stem
        stem_channels = 64
        x = apply_conv2d_block(
            x, stem_channels, 7, 2, activation="relu", name="conv_stem"
        )
        features["STEM_S2"] = x

        # max pooling
        x = layers.ZeroPadding2D(padding=1)(x)
        x = layers.MaxPooling2D(3, strides=2)(x)

        # stages
        output_channels = [64, 128, 256, 512]
        current_stride = 4
        for current_stage_idx, (c, n) in enumerate(
            zip(output_channels, num_blocks)
        ):
            stride = 1 if current_stage_idx == 0 else 2
            current_stride *= stride
            # blocks
            for current_block_idx in range(n):
                stride = stride if current_block_idx == 0 else 1
                name = f"layer{current_stage_idx + 1}_{current_block_idx}"
                if block_fn == "basic":
                    x = apply_basic_block(x, c, stride, name=name)
                elif block_fn == "bottleneck":
                    x = apply_bottleneck_block(x, c, stride, name=name)
                else:
                    raise NotImplementedError
            # add feature
            features[f"BLOCK{current_stage_idx}_S{current_stride}"] = x

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.block_fn = block_fn
        self.num_blocks = num_blocks

    def get_config(self):
        config = super().get_config()
        config.update(
            {"block_fn": self.block_fn, "num_blocks": self.num_blocks}
        )
        return config

    def fix_config(self, config):
        unused_kwargs = ["block_fn", "num_blocks"]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class ResNet18(ResNet):
    available_weights = [
        (
            "imagenet",
            ResNet.default_origin,
            "resnet18_resnet18.a1_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ResNet18",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "basic",
            [2, 2, 2, 2],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ResNet34(ResNet):
    available_weights = [
        (
            "imagenet",
            ResNet.default_origin,
            "resnet34_resnet34.a1_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ResNet34",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "basic",
            [3, 4, 6, 3],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ResNet50(ResNet):
    available_weights = [
        (
            "imagenet",
            ResNet.default_origin,
            "resnet50_resnet50.a1_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ResNet50",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "bottleneck",
            [3, 4, 6, 3],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ResNet101(ResNet):
    available_weights = [
        (
            "imagenet",
            ResNet.default_origin,
            "resnet101_resnet101.a1_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ResNet101",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "bottleneck",
            [3, 4, 23, 3],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ResNet152(ResNet):
    available_weights = [
        (
            "imagenet",
            ResNet.default_origin,
            "resnet152_resnet152.a1_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ResNet152",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "bottleneck",
            [3, 8, 36, 3],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(ResNet18, "imagenet")
add_model_to_registry(ResNet34, "imagenet")
add_model_to_registry(ResNet50, "imagenet")
add_model_to_registry(ResNet101, "imagenet")
add_model_to_registry(ResNet152, "imagenet")



import pathlib
import typing

from keras import backend
from keras import layers
from keras import models
from keras import ops

from kimm.models import BaseModel
from kimm.utils.module_utils import torch


def export_onnx(
    model: BaseModel,
    input_shape: typing.Union[int, typing.Sequence[int]],
    export_path: typing.Union[str, pathlib.Path],
    batch_size: int = 1,
):
    """Export the model to onnx format (in float32).

    Only torch backend with 'channels_first' is supported. The onnx model will
    be generated using `torch.onnx.export` and optimized through `onnxsim` and
    `onnxoptimizer`.

    Note that `onnx`, `onnxruntime`, `onnxsim` and `onnxoptimizer` must be
    installed.

    Args:
        model: keras.Model, the model to be exported.
        input_shape: int or sequence of int, specifying the shape of the input.
        export_path: str or pathlib.Path, specifying the path to export.
        batch_size: int, specifying the batch size of the input,
            defaults to `1`.
    """
    if backend.backend() != "torch":
        raise ValueError("`export_onnx` only supports torch backend")
    if backend.image_data_format() != "channels_first":
        raise ValueError(
            "`export_onnx` only supports 'channels_first' data format."
        )
    try:
        import onnx
        import onnxoptimizer
        import onnxsim
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Failed to import 'onnx', 'onnxsim' or 'onnxoptimizer'. "
            "Please install them by the following instruction:\n"
            "'pip install torch onnx onnxsim onnxoptimizer'"
        )

    if isinstance(input_shape, int):
        input_shape = [3, input_shape, input_shape]
    elif len(input_shape) == 2:
        input_shape = [3, input_shape[0], input_shape[1]]
    elif len(input_shape) == 3:
        input_shape = input_shape

    # Fix input shape
    inputs = layers.Input(
        shape=input_shape, batch_size=batch_size, name="inputs"
    )
    outputs = model(inputs, training=False)
    model = models.Model(inputs, outputs)
    model = model.eval()

    full_input_shape = [1] + list(input_shape)
    dummy_inputs = ops.ones(full_input_shape, dtype="float32")
    scripted_model = torch.jit.trace(
        model.forward, example_inputs=[dummy_inputs]
    )
    torch.onnx.export(scripted_model, dummy_inputs, export_path)

    # Further optimization
    model = onnx.load(export_path)
    model_simp, _ = onnxsim.simplify(model)
    model_simp = onnxoptimizer.optimize(model_simp)
    onnx.save(model_simp, export_path)



import pytest
from absl.testing import parameterized
from keras import backend
from keras.src import testing

from kimm import export
from kimm import models


class ExportOnnxTest(testing.TestCase, parameterized.TestCase):
    def get_model(self):
        input_shape = [3, 224, 224]  # channels_first
        model = models.MobileNet050V3Small(include_preprocessing=False)
        return input_shape, model

    @classmethod
    def setUpClass(cls):
        cls.original_image_data_format = backend.image_data_format()

    @classmethod
    def tearDownClass(cls):
        backend.set_image_data_format(cls.original_image_data_format)

    @pytest.mark.skipif(
        backend.backend() != "torch", reason="Requires torch backend."
    )
    def DISABLE_test_export_onnx_use(self):
        # TODO: turn on this test
        # SystemError: <method '__int__' of 'torch._C._TensorBase' objects>
        # returned a result with an exception set
        backend.set_image_data_format("channels_first")
        input_shape, model = self.get_model()

        temp_dir = self.get_temp_dir()

        export.export_onnx(model, input_shape, f"{temp_dir}/model.onnx")



import pytest
from absl.testing import parameterized
from keras import backend
from keras import ops
from keras import random
from keras.src import testing

from kimm import export
from kimm import models


class ExportTFLiteTest(testing.TestCase, parameterized.TestCase):
    def get_model_and_representative_dataset(self):
        input_shape = [224, 224, 3]
        model = models.MobileNet050V3Small(include_preprocessing=False)

        def representative_dataset():
            for _ in range(10):
                yield [
                    ops.convert_to_numpy(
                        random.uniform([1, *input_shape], maxval=255.0)
                    )
                ]

        return input_shape, model, representative_dataset

    @classmethod
    def setUpClass(cls):
        cls.original_image_data_format = backend.image_data_format()

    @classmethod
    def tearDownClass(cls):
        backend.set_image_data_format(cls.original_image_data_format)

    @pytest.mark.skipif(
        backend.backend() != "tensorflow", reason="Requires tensorflow backend."
    )
    def test_export_tflite_fp32(self):
        (input_shape, model, _) = self.get_model_and_representative_dataset()
        temp_dir = self.get_temp_dir()

        export.export_tflite(
            model, input_shape, f"{temp_dir}/model_fp32.onnx", "float32"
        )

    @pytest.mark.skipif(
        backend.backend() != "tensorflow", reason="Requires tensorflow backend."
    )
    def test_export_tflite_fp16(self):
        (input_shape, model, _) = self.get_model_and_representative_dataset()
        temp_dir = self.get_temp_dir()

        export.export_tflite(
            model, input_shape, f"{temp_dir}/model_fp16.tflite", "float16"
        )

    @pytest.mark.skipif(
        backend.backend() != "tensorflow", reason="Requires tensorflow backend."
    )
    def test_export_tflite_int8(self):
        (
            input_shape,
            model,
            representative_dataset,
        ) = self.get_model_and_representative_dataset()
        temp_dir = self.get_temp_dir()

        export.export_tflite(
            model,
            input_shape,
            f"{temp_dir}/model_int8.tflite",
            "int8",
            representative_dataset,
        )



import keras
from keras import layers
from keras import ops


@keras.saving.register_keras_serializable(package="kimm")
class Attention(layers.Layer):
    def __init__(
        self,
        hidden_dim: int,
        num_heads: int = 8,
        use_qkv_bias: bool = False,
        use_qk_norm: bool = False,
        attention_dropout_rate: float = 0.0,
        projection_dropout_rate: float = 0.0,
        name: str = "attention",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.hidden_dim = hidden_dim
        self.num_heads = num_heads
        self.head_dim = hidden_dim // num_heads
        self.scale = self.head_dim ** (-0.5)
        self.use_qkv_bias = use_qkv_bias
        self.use_qk_norm = use_qk_norm
        self.attention_dropout_rate = attention_dropout_rate
        self.projection_dropout_rate = projection_dropout_rate
        self.name = name

        self.qkv = layers.Dense(
            hidden_dim * 3,
            use_bias=use_qkv_bias,
            dtype=self.dtype,
            name=f"{name}_qkv",
        )
        if use_qk_norm:
            self.q_norm = layers.LayerNormalization(
                dtype=self.dtype_policy, name=f"{name}_q_norm"
            )
            self.k_norm = layers.LayerNormalization(
                dtype=self.dtype_policy, name=f"{name}_k_norm"
            )
        else:
            self.q_norm = layers.Identity(dtype=self.dtype_policy)
            self.k_norm = layers.Identity(dtype=self.dtype_policy)

        self.attention_dropout = layers.Dropout(
            attention_dropout_rate,
            dtype=self.dtype_policy,
            name=f"{name}_attn_drop",
        )
        self.projection = layers.Dense(
            hidden_dim, dtype=self.dtype_policy, name=f"{name}_proj"
        )
        self.projection_dropout = layers.Dropout(
            projection_dropout_rate,
            dtype=self.dtype_policy,
            name=f"{name}_proj_drop",
        )

    def build(self, input_shape):
        self.qkv.build(input_shape)
        qkv_output_shape = list(input_shape)
        qkv_output_shape[-1] = qkv_output_shape[-1] * 3
        self.q_norm.build(qkv_output_shape)
        self.k_norm.build(qkv_output_shape)
        attention_input_shape = [
            input_shape[0],
            self.num_heads,
            input_shape[1],
            input_shape[1],
        ]
        self.attention_dropout.build(attention_input_shape)
        self.projection.build(input_shape)
        self.projection_dropout.build(input_shape)
        self.built = True

    def call(self, inputs, training=None, mask=None):
        input_shape = ops.shape(inputs)
        qkv = self.qkv(inputs)
        qkv = ops.reshape(
            qkv,
            [
                input_shape[0],
                input_shape[1],
                3,
                self.num_heads,
                self.head_dim,
            ],
        )
        qkv = ops.transpose(qkv, [2, 0, 3, 1, 4])
        q, k, v = ops.unstack(qkv, 3, axis=0)
        q = self.q_norm(q)
        k = self.k_norm(k)

        # attention
        q = ops.multiply(q, self.scale)
        attn = ops.matmul(q, ops.swapaxes(k, -2, -1))
        attn = ops.softmax(attn)
        attn = self.attention_dropout(attn)
        x = ops.matmul(attn, v)

        x = ops.swapaxes(x, 1, 2)
        x = ops.reshape(x, input_shape)
        x = self.projection(x)
        x = self.projection_dropout(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "num_heads": self.num_heads,
                "use_qkv_bias": self.use_qkv_bias,
                "use_qk_norm": self.use_qk_norm,
                "attention_dropout_rate": self.attention_dropout_rate,
                "projection_dropout_rate": self.projection_dropout_rate,
                "name": self.name,
            }
        )
        return config



import keras
import pytest
import tensorflow as tf
from absl.testing import parameterized
from keras import backend
from keras import models
from keras import ops
from keras import random
from keras import utils
from keras.applications.imagenet_utils import decode_predictions
from keras.src import testing

from kimm import models as kimm_models
from kimm.utils import make_divisible

# name, class, default_size, features (name, shape),
# weights (defaults to imagenet)
MODEL_CONFIGS = [
    # convmixer
    (
        kimm_models.ConvMixer736D32.__name__,
        kimm_models.ConvMixer736D32,
        224,
        [
            ("STEM", [1, 32, 32, 768]),
            *((f"BLOCK{i}", [1, 32, 32, 768]) for i in range(32)),
        ],
    ),
    # convnext
    (
        kimm_models.ConvNeXtAtto.__name__,
        kimm_models.ConvNeXtAtto,
        288,
        [
            ("STEM_S4", [1, 72, 72, 40]),
            ("BLOCK0_S4", [1, 72, 72, 40]),
            ("BLOCK1_S8", [1, 36, 36, 80]),
            ("BLOCK2_S16", [1, 18, 18, 160]),
            ("BLOCK3_S32", [1, 9, 9, 320]),
        ],
    ),
    # densenet
    (
        kimm_models.DenseNet121.__name__,
        kimm_models.DenseNet121,
        224,
        [
            ("STEM_S4", [1, 56, 56, 64]),
            ("BLOCK0_S8", [1, 28, 28, 128]),
            ("BLOCK1_S16", [1, 14, 14, 256]),
            ("BLOCK2_S32", [1, 7, 7, 512]),
            ("BLOCK3_S32", [1, 7, 7, 1024]),
        ],
    ),
    # efficientnet
    (
        kimm_models.EfficientNetB2.__name__,
        kimm_models.EfficientNetB2,
        260,
        [
            ("STEM_S2", [1, 130, 130, make_divisible(32 * 1.1)]),
            ("BLOCK1_S4", [1, 65, 65, make_divisible(24 * 1.1)]),
            ("BLOCK2_S8", [1, 33, 33, make_divisible(40 * 1.1)]),
            ("BLOCK3_S16", [1, 17, 17, make_divisible(80 * 1.1)]),
            ("BLOCK5_S32", [1, 9, 9, make_divisible(192 * 1.1)]),
        ],
    ),
    (
        kimm_models.EfficientNetLiteB2.__name__,
        kimm_models.EfficientNetLiteB2,
        260,
        [
            ("STEM_S2", [1, 130, 130, make_divisible(32 * 1.1)]),
            ("BLOCK1_S4", [1, 65, 65, make_divisible(24 * 1.1)]),
            ("BLOCK2_S8", [1, 33, 33, make_divisible(40 * 1.1)]),
            ("BLOCK3_S16", [1, 17, 17, make_divisible(80 * 1.1)]),
            ("BLOCK5_S32", [1, 9, 9, make_divisible(192 * 1.1)]),
        ],
    ),
    (
        kimm_models.EfficientNetV2S.__name__,
        kimm_models.EfficientNetV2S,
        300,
        [
            ("STEM_S2", [1, 150, 150, make_divisible(24 * 1.0)]),
            ("BLOCK1_S4", [1, 75, 75, make_divisible(48 * 1.0)]),
            ("BLOCK2_S8", [1, 38, 38, make_divisible(64 * 1.0)]),
            ("BLOCK3_S16", [1, 19, 19, make_divisible(128 * 1.0)]),
            ("BLOCK5_S32", [1, 10, 10, make_divisible(256 * 1.0)]),
        ],
    ),
    (
        kimm_models.EfficientNetV2B0.__name__,
        kimm_models.EfficientNetV2B0,
        192,
        [
            ("STEM_S2", [1, 96, 96, make_divisible(32 * 1.0)]),
            ("BLOCK1_S4", [1, 48, 48, make_divisible(32 * 1.0)]),
            ("BLOCK2_S8", [1, 24, 24, make_divisible(48 * 1.0)]),
            ("BLOCK3_S16", [1, 12, 12, make_divisible(96 * 1.0)]),
            ("BLOCK5_S32", [1, 6, 6, make_divisible(192 * 1.0)]),
        ],
    ),
    (
        kimm_models.TinyNetE.__name__,
        kimm_models.TinyNetE,
        106,
        [
            ("STEM_S2", [1, 53, 53, 32]),
            ("BLOCK1_S4", [1, 27, 27, make_divisible(24 * 0.51)]),
            ("BLOCK2_S8", [1, 14, 14, make_divisible(40 * 0.51)]),
            ("BLOCK3_S16", [1, 7, 7, make_divisible(80 * 0.51)]),
            ("BLOCK5_S32", [1, 4, 4, make_divisible(192 * 0.51)]),
        ],
    ),
    # ghostnet
    (
        kimm_models.GhostNet100.__name__,
        kimm_models.GhostNet100,
        224,
        [
            ("STEM_S2", [1, 112, 112, 16]),
            ("BLOCK1_S4", [1, 56, 56, 24]),
            ("BLOCK3_S8", [1, 28, 28, 40]),
            ("BLOCK5_S16", [1, 14, 14, 80]),
            ("BLOCK7_S32", [1, 7, 7, 160]),
        ],
    ),
    (
        kimm_models.GhostNet100V2.__name__,
        kimm_models.GhostNet100V2,
        224,
        [
            ("STEM_S2", [1, 112, 112, 16]),
            ("BLOCK1_S4", [1, 56, 56, 24]),
            ("BLOCK3_S8", [1, 28, 28, 40]),
            ("BLOCK5_S16", [1, 14, 14, 80]),
            ("BLOCK7_S32", [1, 7, 7, 160]),
        ],
    ),
    # inception_v3
    (
        kimm_models.InceptionV3.__name__,
        kimm_models.InceptionV3,
        299,
        [
            ("STEM_S2", [1, 147, 147, 64]),
            ("BLOCK0_S4", [1, 71, 71, 192]),
            ("BLOCK1_S8", [1, 35, 35, 288]),
            ("BLOCK2_S16", [1, 17, 17, 768]),
            ("BLOCK3_S32", [1, 8, 8, 2048]),
        ],
    ),
    # mobilenet_v2
    (
        kimm_models.MobileNet050V2.__name__,
        kimm_models.MobileNet050V2,
        224,
        [
            ("STEM_S2", [1, 112, 112, make_divisible(32 * 0.5)]),
            ("BLOCK1_S4", [1, 56, 56, make_divisible(24 * 0.5)]),
            ("BLOCK2_S8", [1, 28, 28, make_divisible(32 * 0.5)]),
            ("BLOCK3_S16", [1, 14, 14, make_divisible(64 * 0.5)]),
            ("BLOCK5_S32", [1, 7, 7, make_divisible(160 * 0.5)]),
        ],
    ),
    (
        kimm_models.MobileNet100V2.__name__,
        kimm_models.MobileNet100V2,
        224,
        [
            ("STEM_S2", [1, 112, 112, make_divisible(32 * 1.0)]),
            ("BLOCK1_S4", [1, 56, 56, make_divisible(24 * 1.0)]),
            ("BLOCK2_S8", [1, 28, 28, make_divisible(32 * 1.0)]),
            ("BLOCK3_S16", [1, 14, 14, make_divisible(64 * 1.0)]),
            ("BLOCK5_S32", [1, 7, 7, make_divisible(160 * 1.0)]),
        ],
    ),
    # mobilenet_v3
    (
        kimm_models.LCNet100.__name__,
        kimm_models.LCNet100,
        224,
        [
            ("STEM_S2", [1, 112, 112, make_divisible(16 * 1.0)]),
            ("BLOCK1_S4", [1, 56, 56, make_divisible(64 * 1.0)]),
            ("BLOCK2_S8", [1, 28, 28, make_divisible(128 * 1.0)]),
            ("BLOCK3_S16", [1, 14, 14, make_divisible(256 * 1.0)]),
            ("BLOCK5_S32", [1, 7, 7, make_divisible(512 * 1.0)]),
        ],
    ),
    (
        kimm_models.MobileNet100V3Large.__name__,
        kimm_models.MobileNet100V3Large,
        224,
        [
            ("STEM_S2", [1, 112, 112, make_divisible(16 * 1.0)]),
            ("BLOCK1_S4", [1, 56, 56, make_divisible(24 * 1.0)]),
            ("BLOCK2_S8", [1, 28, 28, make_divisible(40 * 1.0)]),
            ("BLOCK3_S16", [1, 14, 14, make_divisible(80 * 1.0)]),
            ("BLOCK5_S32", [1, 7, 7, make_divisible(160 * 1.0)]),
        ],
    ),
    (
        kimm_models.MobileNet100V3Small.__name__,
        kimm_models.MobileNet100V3Small,
        224,
        [
            ("STEM_S2", [1, 112, 112, make_divisible(16 * 1.0)]),
            ("BLOCK0_S4", [1, 56, 56, make_divisible(16 * 1.0)]),
            ("BLOCK1_S8", [1, 28, 28, make_divisible(24 * 1.0)]),
            ("BLOCK2_S16", [1, 14, 14, make_divisible(40 * 1.0)]),
            ("BLOCK4_S32", [1, 7, 7, make_divisible(96 * 1.0)]),
        ],
    ),
    (
        kimm_models.MobileNet100V3SmallMinimal.__name__,
        kimm_models.MobileNet100V3SmallMinimal,
        224,
        [
            ("STEM_S2", [1, 112, 112, make_divisible(16 * 1.0)]),
            ("BLOCK0_S4", [1, 56, 56, make_divisible(16 * 1.0)]),
            ("BLOCK1_S8", [1, 28, 28, make_divisible(24 * 1.0)]),
            ("BLOCK2_S16", [1, 14, 14, make_divisible(40 * 1.0)]),
            ("BLOCK4_S32", [1, 7, 7, make_divisible(96 * 1.0)]),
        ],
    ),
    # mobilevit
    (
        kimm_models.MobileViTS.__name__,
        kimm_models.MobileViTS,
        256,
        [
            ("STEM_S2", [1, 128, 128, 16]),
            ("BLOCK1_S4", [1, 64, 64, 64]),
            ("BLOCK2_S8", [1, 32, 32, 96]),
            ("BLOCK3_S16", [1, 16, 16, 128]),
            ("BLOCK4_S32", [1, 8, 8, 160]),
        ],
    ),
    (
        kimm_models.MobileViTXS.__name__,
        kimm_models.MobileViTXS,
        256,
        [
            ("STEM_S2", [1, 128, 128, 16]),
            ("BLOCK1_S4", [1, 64, 64, 48]),
            ("BLOCK2_S8", [1, 32, 32, 64]),
            ("BLOCK3_S16", [1, 16, 16, 80]),
            ("BLOCK4_S32", [1, 8, 8, 96]),
        ],
    ),
    # regnet
    (
        kimm_models.RegNetX002.__name__,
        kimm_models.RegNetX002,
        224,
        [
            ("STEM_S2", [1, 112, 112, 32]),
            ("BLOCK0_S4", [1, 56, 56, 24]),
            ("BLOCK1_S8", [1, 28, 28, 56]),
            ("BLOCK2_S16", [1, 14, 14, 152]),
            ("BLOCK3_S32", [1, 7, 7, 368]),
        ],
    ),
    (
        kimm_models.RegNetY002.__name__,
        kimm_models.RegNetY002,
        224,
        [
            ("STEM_S2", [1, 112, 112, 32]),
            ("BLOCK0_S4", [1, 56, 56, 24]),
            ("BLOCK1_S8", [1, 28, 28, 56]),
            ("BLOCK2_S16", [1, 14, 14, 152]),
            ("BLOCK3_S32", [1, 7, 7, 368]),
        ],
    ),
    # resnet
    (
        kimm_models.ResNet18.__name__,
        kimm_models.ResNet18,
        224,
        [
            ("STEM_S2", [1, 112, 112, 64]),
            ("BLOCK0_S4", [1, 56, 56, 64]),
            ("BLOCK1_S8", [1, 28, 28, 128]),
            ("BLOCK2_S16", [1, 14, 14, 256]),
            ("BLOCK3_S32", [1, 7, 7, 512]),
        ],
    ),
    (
        kimm_models.ResNet50.__name__,
        kimm_models.ResNet50,
        224,
        [
            ("STEM_S2", [1, 112, 112, 64]),
            ("BLOCK0_S4", [1, 56, 56, 64 * 4]),
            ("BLOCK1_S8", [1, 28, 28, 128 * 4]),
            ("BLOCK2_S16", [1, 14, 14, 256 * 4]),
            ("BLOCK3_S32", [1, 7, 7, 512 * 4]),
        ],
    ),
    # vgg
    (
        kimm_models.VGG11.__name__,
        kimm_models.VGG11,
        224,
        [
            ("BLOCK0_S1", [1, 224, 224, 64]),
            ("BLOCK1_S2", [1, 112, 112, 128]),
            ("BLOCK2_S4", [1, 56, 56, 256]),
            ("BLOCK3_S8", [1, 28, 28, 512]),
            ("BLOCK4_S16", [1, 14, 14, 512]),
            ("BLOCK5_S32", [1, 7, 7, 512]),
        ],
        None,  # skip weights to save time
    ),
    # vision_transformer
    (
        kimm_models.VisionTransformerTiny16.__name__,
        kimm_models.VisionTransformerTiny16,
        384,
        [*((f"BLOCK{i}", [1, 577, 192]) for i in range(5))],
    ),
    (
        kimm_models.VisionTransformerTiny32.__name__,
        kimm_models.VisionTransformerTiny32,
        384,
        [*((f"BLOCK{i}", [1, 145, 192]) for i in range(5))],
        None,  # no weights
    ),
    # xception
    (
        kimm_models.Xception.__name__,
        kimm_models.Xception,
        299,
        [
            ("STEM_S2", [1, 147, 147, 64]),
            ("BLOCK0_S4", [1, 74, 74, 128]),
            ("BLOCK1_S8", [1, 37, 37, 256]),
            ("BLOCK2_S16", [1, 19, 19, 728]),
            ("BLOCK3_S32", [1, 10, 10, 2048]),
        ],
    ),
]


class ModelTest(testing.TestCase, parameterized.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.original_image_data_format = backend.image_data_format()

    @classmethod
    def tearDownClass(cls):
        backend.set_image_data_format(cls.original_image_data_format)

    @parameterized.named_parameters(MODEL_CONFIGS)
    def test_model_base_channels_last(
        self, model_class, image_size, features, weights="imagenet"
    ):
        backend.set_image_data_format("channels_last")
        model = model_class(weights=weights)
        image_path = keras.utils.get_file(
            "african_elephant.jpg", "https://i.imgur.com/Bvro0YD.png"
        )
        # preprocessing
        image = utils.load_img(image_path, target_size=(image_size, image_size))
        image = utils.img_to_array(image, data_format="channels_last")
        x = ops.convert_to_tensor(image)
        x = ops.expand_dims(x, axis=0)

        y = model(x, training=False)

        if weights == "imagenet":
            names = [p[1] for p in decode_predictions(y)[0]]
            # Test correct label is in top 3 (weak correctness test).
            self.assertIn("African_elephant", names[:3])
        elif weights is None:
            self.assertEqual(list(y.shape), [1, 1000])

    @parameterized.named_parameters(MODEL_CONFIGS)
    def test_model_base_channels_first(
        self, model_class, image_size, features, weights="imagenet"
    ):
        if (
            len(tf.config.list_physical_devices("GPU")) == 0
            and backend.backend() == "tensorflow"
        ):
            self.skipTest(
                "Conv2D doesn't support channels_first using CPU with "
                "tensorflow backend"
            )

        backend.set_image_data_format("channels_first")
        model = model_class(weights=weights)
        image_path = keras.utils.get_file(
            "african_elephant.jpg", "https://i.imgur.com/Bvro0YD.png"
        )
        # preprocessing
        image = utils.load_img(image_path, target_size=(image_size, image_size))
        image = utils.img_to_array(image, data_format="channels_first")
        x = ops.convert_to_tensor(image)
        x = ops.expand_dims(x, axis=0)

        y = model(x, training=False)

        if weights == "imagenet":
            names = [p[1] for p in decode_predictions(y)[0]]
            # Test correct label is in top 3 (weak correctness test).
            self.assertIn("African_elephant", names[:3])
        elif weights is None:
            self.assertEqual(list(y.shape), [1, 1000])

    @parameterized.named_parameters(MODEL_CONFIGS)
    def test_model_feature_extractor(
        self, model_class, image_size, features, weights="imagenet"
    ):
        backend.set_image_data_format("channels_last")
        x = random.uniform([1, image_size, image_size, 3]) * 255.0
        model = model_class(weights=None, feature_extractor=True)

        y = model(x, training=False)

        self.assertIsInstance(y, dict)
        self.assertContainsSubset(
            model_class.available_feature_keys, list(y.keys())
        )
        for feature_info in features:
            name, shape = feature_info
            self.assertEqual(list(y[name].shape), shape)

    @pytest.mark.serialization
    @parameterized.named_parameters(MODEL_CONFIGS)
    def test_model_serialization(
        self, model_class, image_size, features, weights="imagenet"
    ):
        backend.set_image_data_format("channels_last")
        x = random.uniform([1, image_size, image_size, 3]) * 255.0
        temp_dir = self.get_temp_dir()
        model1 = model_class(weights=None)

        y1 = model1(x, training=False)
        model1.save(temp_dir + "/model.keras")
        model2 = models.load_model(temp_dir + "/model.keras")
        y2 = model2(x, training=False)

        self.assertAllClose(y1, y2)



from absl.testing import parameterized
from keras import layers
from keras import random
from keras.src import testing

from kimm.models.base_model import BaseModel


class SampleModel(BaseModel):
    def __init__(self, **kwargs):
        self.set_properties(kwargs)
        inputs = layers.Input(shape=[224, 224, 3])

        features = {}
        s2 = layers.Conv2D(3, 1, 2, use_bias=False)(inputs)
        features["S2"] = s2
        s4 = layers.Conv2D(3, 1, 2, use_bias=False)(s2)
        features["S4"] = s4
        s8 = layers.Conv2D(3, 1, 2, use_bias=False)(s4)
        features["S8"] = s8
        s16 = layers.Conv2D(3, 1, 2, use_bias=False)(s8)
        features["S16"] = s16
        s32 = layers.Conv2D(3, 1, 2, use_bias=False)(s16)
        features["S32"] = s32
        super().__init__(
            inputs=inputs, outputs=s32, features=features, **kwargs
        )

    @staticmethod
    def available_feature_keys():
        # predefined for better UX
        return [f"S{2**i}" for i in range(1, 6)]

    def get_config(self):
        return super().get_config()


class BaseModelTest(testing.TestCase, parameterized.TestCase):
    def test_feature_extractor(self):
        x = random.uniform([1, 224, 224, 3])

        # availiable_feature_keys
        self.assertContainsSubset(
            ["S2", "S4", "S8", "S16", "S32"],
            SampleModel.available_feature_keys(),
        )

        # feature_extractor=False
        model = SampleModel()

        y = model(x, training=False)

        self.assertNotIsInstance(y, dict)
        self.assertEqual(list(y.shape), [1, 7, 7, 3])

        # feature_extractor=True
        model = SampleModel(feature_extractor=True)

        y = model(x, training=False)

        self.assertIsInstance(y, dict)
        self.assertEqual(list(y["S2"].shape), [1, 112, 112, 3])
        self.assertEqual(list(y["S32"].shape), [1, 7, 7, 3])

        # feature_extractor=True with feature_keys
        model = SampleModel(
            feature_extractor=True, feature_keys=["S2", "S16", "S32"]
        )

        y = model(x, training=False)

        self.assertIsInstance(y, dict)
        self.assertNotIn("S4", y)
        self.assertNotIn("S8", y)
        self.assertEqual(list(y["S2"].shape), [1, 112, 112, 3])
        self.assertEqual(list(y["S16"].shape), [1, 14, 14, 3])
        self.assertEqual(list(y["S32"].shape), [1, 7, 7, 3])
